local iqm = require("libs.iqm")
local cpml = require("libs.cpml")
local functions = {}

-- Globals to local
local math = math
local love = love
local os = os
local ipairs = ipairs
local table = table
local tostring = tostring
local type = type
local pairs = pairs
local image = image

function functions.round(num)
    local under = math.floor(num)
    local upper = math.floor(num) + 1
    local underV = -(under - num)
    local upperV = upper - num
    if (upperV > underV) then
        return under
    else
        return upper
    end
end
-- generate assets (see love.load)
-- generate sounds
function functions.generateClickySound()
    local snd = love.sound.newSoundData(512, 44100, 16, 1)
    for i = 0,snd:getSampleCount()-1 do
        local t = i / 44100
        local s = i / snd:getSampleCount()
        snd:setSample(i, (.7*(2*love.math.random()-1) + .3*math.sin(t*9000*math.pi)) * (1-s)^1.2 * .3)
    end
    return love.audio.newSource(snd)
end

function functions.SecondsToClock(sSeconds)
	local nSeconds = tonumber(sSeconds)
	if nSeconds == 0 then
		--return nil;
		return "00:00:00";
	else
		nHours = string.format("%02.f", math.floor(nSeconds/3600));
		nMins = string.format("%02.f", math.floor(nSeconds/60 - (nHours*60)));
		nSecs = string.format("%02.f", math.floor(nSeconds - nHours*3600 - nMins *60));
		return nHours..":"..nMins..":"..nSecs
	end
end

function functions.generateBox(width,height,color, effect)
  if not effect then
    effect = 0
  end
  if not color then
    color = {255,255,255}
  end
  local alpha = color[4] or 255
  local vertices = {
                {
                        -- top-left corner (red-tinted)
                        0, 3, -- position of the vertex
                        0, 0, -- texture coordinate at the vertex position
                        color[1], color[2], color[3], alpha -- color of the vertex
                },
                {
                        -- top-left corner (red-tinted)
                        3, 0, -- position of the vertex
                        0, 0, -- texture coordinate at the vertex position
                        color[1], color[2], color[3], alpha -- color of the vertex
                },
                {
                        -- top-right corner (green-tinted)
                        width-3, 0,
                        1, 0, -- texture coordinates are in the range of [0, 1]
                        color[1], color[2], color[3], alpha -- color of the vertex
                },
                {
                        -- top-right corner (green-tinted)
                        width, 3,
                        1, 0, -- texture coordinates are in the range of [0, 1]
                        color[1], color[2], color[3], alpha -- color of the vertex
                },
                {
                        -- bottom-right corner (blue-tinted)
                        width, height-3,
                        1, 1,
                        color[1]+effect, color[2]+effect, color[3]+effect, alpha -- color of the vertex
                },
                {
                        -- bottom-right corner (blue-tinted)
                        width-3, height,
                        1, 1,
                        color[1]+effect, color[2]+effect, color[3]+effect, alpha -- color of the vertex
                },
                {
                        -- bottom-left corner (yellow-tinted)
                        3, height,
                        0, 1,
                        color[1]+effect, color[2]+effect, color[3]+effect, alpha -- color of the vertex
                },
                {
                        -- bottom-left corner (yellow-tinted)
                        0, height-3,
                        0, 1,
                        color[1]+effect, color[2]+effect, color[3]+effect, alpha -- color of the vertex
                },
        }
        -- the Mesh DrawMode "fan" works well for 4-vertex Meshes.
        local mesh = love.graphics.newMesh(vertices, "fan")
        mesh:setTexture(image)
        return mesh
end
function functions.GetFileExtension(url)
  return url:match("^.+(%..+)$")
end
function functions.GetFileName(url)
  local extension = url:match "[^.]+$"
  local tmp = url:gsub("."..extension, "")
  return tmp
end

function functions.split(pString, pPattern)
   local Table = {}  -- NOTE: use {n = 0} in Lua-5.0
   local fpat = "(.-)" .. pPattern
   local last_end = 1
   local s, e, cap = pString:find(fpat, 1)
   while s do
      if s ~= 1 or cap ~= "" then
     table.insert(Table,cap)
      end
      last_end = e+1
      s, e, cap = pString:find(fpat, last_end)
   end
   if last_end <= #pString then
      cap = pString:sub(last_end)
      table.insert(Table, cap)
   end
   return Table
end

function functions.recursiveEnumerate(folder)
        local tabulka = {}
        local lfs = love.filesystem
        local filesTable = lfs.getDirectoryItems(folder)
        for i,v in ipairs(filesTable) do
                local file = folder.."/"..v
                if lfs.isFile(file) then
                        print("Loading: "..v)
                        if functions.GetFileExtension(v) == ".png" or
                        functions.GetFileExtension(v) == ".jpg" or
                        functions.GetFileExtension(v) == ".bmp" then
                          tabulka[functions.GetFileName(v)] = love.graphics.newImage(file, { mipmaps = true })
                        elseif functions.GetFileExtension(v) == ".iqm" then
                          tabulka[functions.GetFileName(v)] = {
                            model = iqm.load(file),
                            position = cpml.vec3(0, 0, 0),
                            orientation = cpml.quat()
                          }
                        else
                          tabulka[functions.GetFileName(v)] = file
                        end
                elseif lfs.isDirectory(file) then
                        tabulka[v] = functions.recursiveEnumerate(file)
                end
        end
        return tabulka
end

function functions.range(from, to, step)
  step = step or 1
  return function(_, lastvalue)
    local nextvalue = lastvalue + step
    if step > 0 and nextvalue <= to or step < 0 and nextvalue >= to or
       step == 0
    then
      return nextvalue
    end
  end, nil, from - step
end

function functions.TableConcat(t1,t2)
    for i=1,#t2 do
        t1[#t1+1] = t2[i]
    end
    return t1
end
function functions.random(min,max)
  if max > min then
    return math.random(min,max)
  else
    return min
  end
end

function functions.tableToText(table)
  local tmp = ""
  for i,v in pairs(table) do
    if type(v) == "table" then
      --tmp = tmp .. functions.tableToText(table)
    else
      tmp = tmp..tostring(i)..":"..tostring(v).."\n"
    end
  end
  return tmp
end

function functions.tableToDict(table)
  local tmp = "{"
  for i,v in pairs(table) do
    if type(v) == "table" then
      --tmp = tmp .. functions.tableToDict(table)
    else
      tmp = tmp.."\'"..tostring(i).."\'"..": "
      if type(v) == "string" then
        tmp = tmp.."\'"..tostring(v).."\'"..","
      else
        tmp = tmp..tostring(v)..","
      end
    end
  end
  tmp = tmp.."}"
  return tmp
end

function functions.tableToValues(table)
  local tmp = ""
  --tmp = "return {"
  for i,v in pairs(table) do
    if type(v) ~= "table" then
      if type(i) == "number" then
        tmp = tmp.."\""..tostring(v).."\",\n"
      else
        tmp = tmp..tostring(i).."= \""..tostring(v).."\",\n"
      end
    else
      if type(i) == "number" then
        tmp = tmp.."{\n"
      else
        tmp = tmp..tostring(i).."= {\n"
      end
      tmp = tmp..tostring(i).."= {\n"
      tmp = tmp..functions.tableToValues(v)
      tmp = tmp.."},\n"
    end
  end
  --tmp = tmp.."}"
  return tmp
end

function functions.recursivelyDelete(item, depth)
        if love.filesystem.isDirectory(item) then
            for _, child in pairs(love.filesystem.getDirectoryItems(item)) do
                functions.recursivelyDelete(item .. '/' .. child, depth + 1);
                love.filesystem.remove(item .. '/' .. child);
            end
        elseif love.filesystem.isFile(item) then
            love.filesystem.remove(item);
        end
        love.filesystem.remove(item)
end

function functions.words(string)
  local words = {}
  for word in string:gmatch("%w+") do table.insert(words, word) end
  return words
end

function functions.saveTable(table)
  local tmp = "return {\n"
  tmp = tmp..functions.tableToValues(table)
  tmp = tmp.."}"
  return tmp
end

function functions.isInside(x1,y1,x2,y2,range, range2)
  range2 = range2 or range
  if x1 <= x2+range and x1 >= x2-range and
    y1 <= y2+range2 and y1 >= y2-range2 then
      return true
  end
  return false
end

function functions.isInside2(x1,y1,x2,y2,x2m,y2m)
  if x1 <= x2m and x1 >= x2 and
    y1 <= y2m and y1 >= y2 then
      return true
  end
  return false
end

function functions.isInside3(obj,min,max,angle)
  local angle = angle or 0
  if obj.x >=min.x and obj.y >= min.y and obj.z >= min.z and
    obj.x <= max.x and obj.y <= max.y and obj.z <= max.z then
      return true
  end
  return false
end

function functions.contain(t,value)
  if not t then return false end
  for i,v in pairs(t) do
    if v == value then return true end
  end
  return false
end

return functions
